@@ -1,13 +1,10 @@
 // Take a look at the license at the top of the repository in the LICENSE file.
 
-use glib::translate::{from_glib, Borrowed, FromGlibPtrNone, ToGlibPtr};
+use std::{fmt, marker::PhantomData, mem, ops, ptr, slice};
 
-use std::fmt;
-use std::marker::PhantomData;
-use std::mem;
-use std::ops;
-use std::ptr;
-use std::slice;
+use glib::translate::*;
+
+use smallvec::SmallVec;
 
 pub enum Readable {}
 pub enum Writable {}
@@ -15,8 +12,7 @@ pub enum Writable {}
 pub struct AudioBuffer<T> {
     // Has to be boxed because it contains self-references
     audio_buffer: Box<ffi::GstAudioBuffer>,
-    buffer: Option<gst::Buffer>,
-    info: crate::AudioInfo,
+    buffer: gst::Buffer,
     phantom: PhantomData<T>,
 }
 
@@ -26,76 +22,101 @@ unsafe impl<T> Sync for AudioBuffer<T> {}
 impl<T> fmt::Debug for AudioBuffer<T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.debug_struct("AudioBuffer")
-            .field("audio_buffer", &self.audio_buffer)
-            .field("buffer", &self.buffer)
-            .field("info", &self.info)
-            .field("phantom", &self.phantom)
+            .field("n_samples", &self.n_samples())
+            .field("n_planes", &self.n_planes())
+            .field("buffer", &self.buffer())
+            .field("info", &self.info())
             .finish()
     }
 }
 
 impl<T> AudioBuffer<T> {
+    #[inline]
     pub fn info(&self) -> &crate::AudioInfo {
-        &self.info
+        unsafe {
+            &*(&self.audio_buffer.info as *const ffi::GstAudioInfo as *const crate::AudioInfo)
+        }
     }
 
-    pub fn into_buffer(mut self) -> gst::Buffer {
-        self.buffer.take().unwrap()
+    #[inline]
+    pub fn into_buffer(self) -> gst::Buffer {
+        unsafe {
+            let mut s = mem::ManuallyDrop::new(self);
+            let buffer = ptr::read(&s.buffer);
+            ffi::gst_audio_buffer_unmap(&mut *s.audio_buffer);
+            ptr::drop_in_place(&mut s.audio_buffer);
+
+            buffer
+        }
     }
 
+    #[inline]
     pub fn format(&self) -> crate::AudioFormat {
         self.info().format()
     }
 
+    #[inline]
     pub fn format_info(&self) -> crate::AudioFormatInfo {
         self.info().format_info()
     }
 
+    #[inline]
     pub fn channels(&self) -> u32 {
         self.info().channels()
     }
 
+    #[inline]
     pub fn rate(&self) -> u32 {
         self.info().rate()
     }
 
+    #[inline]
     pub fn layout(&self) -> crate::AudioLayout {
         self.info().layout()
     }
 
+    #[inline]
     pub fn width(&self) -> u32 {
         self.info().width()
     }
 
+    #[inline]
     pub fn depth(&self) -> u32 {
         self.info().depth()
     }
 
+    #[inline]
     pub fn sample_stride(&self) -> u32 {
         self.info().width() / 8
     }
 
+    #[inline]
     pub fn bps(&self) -> u32 {
         self.info().bps()
     }
 
+    #[inline]
     pub fn bpf(&self) -> u32 {
         self.info().bpf()
     }
 
+    #[inline]
     pub fn n_samples(&self) -> usize {
         self.audio_buffer.n_samples
     }
 
+    #[inline]
     pub fn n_planes(&self) -> u32 {
         self.audio_buffer.n_planes as u32
     }
 
+    #[inline]
     pub fn plane_size(&self) -> usize {
-        (self.n_samples() as usize * self.sample_stride() as usize * self.channels() as usize)
+        (self.n_samples() * self.sample_stride() as usize * self.channels() as usize)
             / self.n_planes() as usize
     }
 
+    #[inline]
     pub fn buffer(&self) -> &gst::BufferRef {
         unsafe { gst::BufferRef::from_ptr(self.audio_buffer.buffer) }
     }
@@ -115,22 +136,33 @@ impl<T> AudioBuffer<T> {
         }
     }
 
+    pub fn planes_data(&self) -> SmallVec<[&[u8]; 8]> {
+        let mut planes = SmallVec::default();
+
+        for plane in 0..self.n_planes() {
+            planes[plane as usize] = self.plane_data(plane).unwrap();
+        }
+
+        planes
+    }
+
+    #[inline]
     pub fn as_audio_buffer_ref(&self) -> AudioBufferRef<&gst::BufferRef> {
-        let info = self.info.clone();
         AudioBufferRef {
             audio_buffer: AudioBufferPtr::Borrowed(ptr::NonNull::from(&*self.audio_buffer)),
-            buffer: Some(self.buffer()),
-            info,
             unmap: false,
+            phantom: PhantomData,
         }
     }
 
+    #[inline]
     pub fn as_ptr(&self) -> *const ffi::GstAudioBuffer {
         &*self.audio_buffer
     }
 }
 
 impl<T> Drop for AudioBuffer<T> {
+    #[inline]
     fn drop(&mut self) {
         unsafe {
             ffi::gst_audio_buffer_unmap(&mut *self.audio_buffer);
@@ -139,6 +171,7 @@ impl<T> Drop for AudioBuffer<T> {
 }
 
 impl AudioBuffer<Readable> {
+    #[inline]
     pub fn from_buffer_readable(
         buffer: gst::Buffer,
         info: &crate::AudioInfo,
@@ -159,21 +192,23 @@ impl AudioBuffer<Readable> {
             if !res {
                 Err(buffer)
             } else {
-                let info = crate::AudioInfo::from_glib_none(
-                    &audio_buffer.info as *const _ as *mut ffi::GstAudioInfo,
-                );
                 Ok(Self {
                     audio_buffer,
-                    buffer: Some(buffer),
-                    info,
+                    buffer,
                     phantom: PhantomData,
                 })
             }
         }
     }
+
+    #[inline]
+    pub fn buffer_owned(&self) -> gst::Buffer {
+        unsafe { from_glib_none(self.audio_buffer.buffer) }
+    }
 }
 
 impl AudioBuffer<Writable> {
+    #[inline]
     pub fn from_buffer_writable(
         buffer: gst::Buffer,
         info: &crate::AudioInfo,
@@ -194,23 +229,15 @@ impl AudioBuffer<Writable> {
             if !res {
                 Err(buffer)
             } else {
-                let info = crate::AudioInfo::from_glib_none(
-                    &audio_buffer.info as *const _ as *mut ffi::GstAudioInfo,
-                );
                 Ok(Self {
                     audio_buffer,
-                    buffer: Some(buffer),
-                    info,
+                    buffer,
                     phantom: PhantomData,
                 })
             }
         }
     }
 
-    pub fn buffer_mut(&mut self) -> &mut gst::BufferRef {
-        unsafe { gst::BufferRef::from_mut_ptr(self.audio_buffer.buffer) }
-    }
-
     pub fn plane_data_mut(&mut self, plane: u32) -> Result<&mut [u8], glib::BoolError> {
         if plane >= self.n_planes() {
             return Err(glib::bool_error!(
@@ -226,16 +253,29 @@ impl AudioBuffer<Writable> {
         }
     }
 
+    pub fn planes_data_mut(&mut self) -> SmallVec<[&mut [u8]; 8]> {
+        let mut planes = SmallVec::default();
+
+        unsafe {
+            for plane in 0..self.n_planes() {
+                let slice = self.plane_data_mut(plane).unwrap();
+                planes.push(slice::from_raw_parts_mut(slice.as_mut_ptr(), slice.len()));
+            }
+        }
+
+        planes
+    }
+
+    #[inline]
     pub fn as_mut_audio_buffer_ref(&mut self) -> AudioBufferRef<&mut gst::BufferRef> {
-        let info = self.info.clone();
         AudioBufferRef {
             audio_buffer: AudioBufferPtr::Borrowed(ptr::NonNull::from(&mut *self.audio_buffer)),
-            buffer: Some(self.buffer_mut()),
-            info,
             unmap: false,
+            phantom: PhantomData,
         }
     }
 
+    #[inline]
     pub fn as_mut_ptr(&mut self) -> *mut ffi::GstAudioBuffer {
         &mut *self.audio_buffer
     }
@@ -250,6 +290,7 @@ enum AudioBufferPtr {
 impl ops::Deref for AudioBufferPtr {
     type Target = ffi::GstAudioBuffer;
 
+    #[inline]
     fn deref(&self) -> &Self::Target {
         match self {
             Self::Owned(ref b) => b,
@@ -259,6 +300,7 @@ impl ops::Deref for AudioBufferPtr {
 }
 
 impl ops::DerefMut for AudioBufferPtr {
+    #[inline]
     fn deref_mut(&mut self) -> &mut Self::Target {
         match self {
             Self::Owned(ref mut b) => &mut *b,
@@ -267,70 +309,97 @@ impl ops::DerefMut for AudioBufferPtr {
     }
 }
 
-#[derive(Debug)]
 pub struct AudioBufferRef<T> {
     // Has to be boxed because it contains self-references
     audio_buffer: AudioBufferPtr,
-    buffer: Option<T>,
-    info: crate::AudioInfo,
     unmap: bool,
+    phantom: PhantomData<T>,
+}
+
+impl<T> fmt::Debug for AudioBufferRef<T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.debug_struct("AudioBufferRef")
+            .field("n_samples", &self.n_samples())
+            .field("n_planes", &self.n_planes())
+            .field("buffer", &unsafe {
+                gst::BufferRef::from_ptr(self.audio_buffer.buffer)
+            })
+            .field("info", &self.info())
+            .finish()
+    }
 }
 
 impl<T> AudioBufferRef<T> {
+    #[inline]
     pub fn info(&self) -> &crate::AudioInfo {
-        &self.info
+        unsafe {
+            &*(&self.audio_buffer.info as *const ffi::GstAudioInfo as *const crate::AudioInfo)
+        }
     }
 
+    #[inline]
     pub fn format(&self) -> crate::AudioFormat {
         self.info().format()
     }
 
+    #[inline]
     pub fn format_info(&self) -> crate::AudioFormatInfo {
         self.info().format_info()
     }
 
+    #[inline]
     pub fn channels(&self) -> u32 {
         self.info().channels()
     }
 
+    #[inline]
     pub fn rate(&self) -> u32 {
         self.info().rate()
     }
 
+    #[inline]
     pub fn layout(&self) -> crate::AudioLayout {
         self.info().layout()
     }
 
+    #[inline]
     pub fn width(&self) -> u32 {
         self.info().width()
     }
 
+    #[inline]
     pub fn depth(&self) -> u32 {
         self.info().depth()
     }
 
+    #[inline]
     pub fn sample_stride(&self) -> u32 {
         self.info().width() / 8
     }
 
+    #[inline]
     pub fn bps(&self) -> u32 {
         self.info().bps()
     }
 
+    #[inline]
     pub fn bpf(&self) -> u32 {
         self.info().bpf()
     }
 
+    #[inline]
     pub fn n_samples(&self) -> usize {
         self.audio_buffer.n_samples
     }
 
+    #[inline]
     pub fn n_planes(&self) -> u32 {
         self.audio_buffer.n_planes as u32
     }
 
+    #[inline]
     pub fn plane_size(&self) -> usize {
-        (self.n_samples() as usize * self.sample_stride() as usize * self.channels() as usize)
+        (self.n_samples() * self.sample_stride() as usize * self.channels() as usize)
             / self.n_planes() as usize
     }
 
@@ -353,29 +422,37 @@ impl<T> AudioBufferRef<T> {
         }
     }
 
+    pub fn planes_data(&self) -> SmallVec<[&[u8]; 8]> {
+        let mut planes = SmallVec::default();
+
+        for plane in 0..self.n_planes() {
+            planes[plane as usize] = self.plane_data(plane).unwrap();
+        }
+
+        planes
+    }
+
+    #[inline]
     pub fn as_ptr(&self) -> *const ffi::GstAudioBuffer {
         &*self.audio_buffer
     }
 }
 
 impl<'a> AudioBufferRef<&'a gst::BufferRef> {
+    #[inline]
     pub unsafe fn from_glib_borrow(audio_buffer: *const ffi::GstAudioBuffer) -> Borrowed<Self> {
-        assert!(!audio_buffer.is_null());
+        debug_assert!(!audio_buffer.is_null());
 
-        let info = crate::AudioInfo::from_glib_none(
-            &(*audio_buffer).info as *const _ as *mut ffi::GstAudioInfo,
-        );
-        let buffer = gst::BufferRef::from_ptr((*audio_buffer).buffer);
         Borrowed::new(Self {
             audio_buffer: AudioBufferPtr::Borrowed(ptr::NonNull::new_unchecked(
                 audio_buffer as *mut _,
             )),
-            buffer: Some(buffer),
-            info,
             unmap: false,
+            phantom: PhantomData,
         })
     }
 
+    #[inline]
     pub fn from_buffer_ref_readable<'b>(
         buffer: &'a gst::BufferRef,
         info: &'b crate::AudioInfo,
@@ -396,40 +473,34 @@ impl<'a> AudioBufferRef<&'a gst::BufferRef> {
             if !res {
                 Err(glib::bool_error!("Failed to map AudioBuffer"))
             } else {
-                let info = crate::AudioInfo::from_glib_none(
-                    &audio_buffer.info as *const _ as *mut ffi::GstAudioInfo,
-                );
                 Ok(Self {
                     audio_buffer: AudioBufferPtr::Owned(audio_buffer),
-                    buffer: Some(buffer),
-                    info,
                     unmap: true,
+                    phantom: PhantomData,
                 })
             }
         }
     }
 
+    #[inline]
     pub fn buffer(&self) -> &gst::BufferRef {
-        self.buffer.as_ref().unwrap()
+        unsafe { gst::BufferRef::from_ptr(self.audio_buffer.buffer) }
     }
 }
 
 impl<'a> AudioBufferRef<&'a mut gst::BufferRef> {
+    #[inline]
     pub unsafe fn from_glib_borrow_mut(audio_buffer: *mut ffi::GstAudioBuffer) -> Borrowed<Self> {
-        assert!(!audio_buffer.is_null());
+        debug_assert!(!audio_buffer.is_null());
 
-        let info = crate::AudioInfo::from_glib_none(
-            &(*audio_buffer).info as *const _ as *mut ffi::GstAudioInfo,
-        );
-        let buffer = gst::BufferRef::from_mut_ptr((*audio_buffer).buffer);
         Borrowed::new(Self {
             audio_buffer: AudioBufferPtr::Borrowed(ptr::NonNull::new_unchecked(audio_buffer)),
-            buffer: Some(buffer),
-            info,
             unmap: false,
+            phantom: PhantomData,
         })
     }
 
+    #[inline]
     pub fn from_buffer_ref_writable<'b>(
         buffer: &'a mut gst::BufferRef,
         info: &'b crate::AudioInfo,
@@ -450,23 +521,16 @@ impl<'a> AudioBufferRef<&'a mut gst::BufferRef> {
             if !res {
                 Err(glib::bool_error!("Failed to map AudioBuffer"))
             } else {
-                let info = crate::AudioInfo::from_glib_none(
-                    &audio_buffer.info as *const _ as *mut ffi::GstAudioInfo,
-                );
                 Ok(Self {
                     audio_buffer: AudioBufferPtr::Owned(audio_buffer),
-                    buffer: Some(buffer),
-                    info,
                     unmap: true,
+                    phantom: PhantomData,
                 })
             }
         }
     }
 
-    pub fn buffer_mut(&mut self) -> &mut gst::BufferRef {
-        self.buffer.as_mut().unwrap()
-    }
-
+    #[inline]
     pub fn plane_data_mut(&mut self, plane: u32) -> Result<&mut [u8], glib::BoolError> {
         if plane >= self.n_planes() {
             return Err(glib::bool_error!(
@@ -486,6 +550,20 @@ impl<'a> AudioBufferRef<&'a mut gst::BufferRef> {
         }
     }
 
+    pub fn planes_data_mut(&mut self) -> SmallVec<[&mut [u8]; 8]> {
+        let mut planes = SmallVec::default();
+
+        unsafe {
+            for plane in 0..self.n_planes() {
+                let slice = self.plane_data_mut(plane).unwrap();
+                planes.push(slice::from_raw_parts_mut(slice.as_mut_ptr(), slice.len()));
+            }
+        }
+
+        planes
+    }
+
+    #[inline]
     pub fn as_mut_ptr(&mut self) -> *mut ffi::GstAudioBuffer {
         &mut *self.audio_buffer
     }
@@ -494,6 +572,7 @@ impl<'a> AudioBufferRef<&'a mut gst::BufferRef> {
 impl<'a> ops::Deref for AudioBufferRef<&'a mut gst::BufferRef> {
     type Target = AudioBufferRef<&'a gst::BufferRef>;
 
+    #[inline]
     fn deref(&self) -> &Self::Target {
         unsafe { &*(self as *const Self as *const Self::Target) }
     }
@@ -503,6 +582,7 @@ unsafe impl<T> Send for AudioBufferRef<T> {}
 unsafe impl<T> Sync for AudioBufferRef<T> {}
 
 impl<T> Drop for AudioBufferRef<T> {
+    #[inline]
     fn drop(&mut self) {
         unsafe {
             if self.unmap {
