@@ -1,11 +1,10 @@
 // Take a look at the license at the top of the repository in the LICENSE file.
 
-use crate::AudioDecoder;
-use crate::AudioInfo;
-use glib::prelude::*;
-use glib::translate::*;
-use std::mem;
-use std::ptr;
+use std::{mem, ptr};
+
+use glib::{prelude::*, translate::*};
+
+use crate::{AudioDecoder, AudioInfo};
 
 extern "C" {
     fn _gst_audio_decoder_error(
@@ -21,78 +20,12 @@ extern "C" {
     ) -> gst::ffi::GstFlowReturn;
 }
 
-pub trait AudioDecoderExtManual: 'static {
-    fn finish_frame(
-        &self,
-        buffer: Option<gst::Buffer>,
-        frames: i32,
-    ) -> Result<gst::FlowSuccess, gst::FlowError>;
-
-    #[cfg(any(feature = "v1_16", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
-    fn finish_subframe(
-        &self,
-        buffer: Option<gst::Buffer>,
-    ) -> Result<gst::FlowSuccess, gst::FlowError>;
-
-    fn negotiate(&self) -> Result<(), gst::FlowError>;
-
-    #[cfg(any(feature = "v1_16", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
-    fn set_output_caps(&self, caps: &gst::Caps) -> Result<(), gst::FlowError>;
-
-    fn set_output_format(&self, info: &AudioInfo) -> Result<(), gst::FlowError>;
-
-    #[doc(alias = "get_allocator")]
-    fn allocator(&self) -> (Option<gst::Allocator>, gst::AllocationParams);
-
-    #[allow(clippy::too_many_arguments)]
-    fn error<T: gst::MessageErrorDomain>(
-        &self,
-        weight: i32,
-        code: T,
-        message: Option<&str>,
-        debug: Option<&str>,
-        file: &str,
-        function: &str,
-        line: u32,
-    ) -> Result<gst::FlowSuccess, gst::FlowError>;
-
-    fn sink_pad(&self) -> &gst::Pad;
-
-    fn src_pad(&self) -> &gst::Pad;
+mod sealed {
+    pub trait Sealed {}
+    impl<T: super::IsA<super::AudioDecoder>> Sealed for T {}
 }
 
-impl<O: IsA<AudioDecoder>> AudioDecoderExtManual for O {
-    fn finish_frame(
-        &self,
-        buffer: Option<gst::Buffer>,
-        frames: i32,
-    ) -> Result<gst::FlowSuccess, gst::FlowError> {
-        unsafe {
-            try_from_glib(ffi::gst_audio_decoder_finish_frame(
-                self.as_ref().to_glib_none().0,
-                buffer.map(|b| b.into_glib_ptr()).unwrap_or(ptr::null_mut()),
-                frames,
-            ))
-        }
-    }
-
-    #[cfg(any(feature = "v1_16", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
-    #[doc(alias = "gst_audio_decoder_finish_subframe")]
-    fn finish_subframe(
-        &self,
-        buffer: Option<gst::Buffer>,
-    ) -> Result<gst::FlowSuccess, gst::FlowError> {
-        unsafe {
-            try_from_glib(ffi::gst_audio_decoder_finish_subframe(
-                self.as_ref().to_glib_none().0,
-                buffer.map(|b| b.into_glib_ptr()).unwrap_or(ptr::null_mut()),
-            ))
-        }
-    }
-
+pub trait AudioDecoderExtManual: sealed::Sealed + IsA<AudioDecoder> + 'static {
     #[doc(alias = "gst_audio_decoder_negotiate")]
     fn negotiate(&self) -> Result<(), gst::FlowError> {
         unsafe {
@@ -107,8 +40,8 @@ impl<O: IsA<AudioDecoder>> AudioDecoderExtManual for O {
         }
     }
 
-    #[cfg(any(feature = "v1_16", feature = "dox"))]
-    #[cfg_attr(feature = "dox", doc(cfg(feature = "v1_16")))]
+    #[cfg(feature = "v1_16")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "v1_16")))]
     #[doc(alias = "gst_audio_decoder_set_output_caps")]
     fn set_output_caps(&self, caps: &gst::Caps) -> Result<(), gst::FlowError> {
         unsafe {
@@ -139,20 +72,22 @@ impl<O: IsA<AudioDecoder>> AudioDecoderExtManual for O {
         }
     }
 
+    #[doc(alias = "get_allocator")]
     #[doc(alias = "gst_audio_decoder_get_allocator")]
     fn allocator(&self) -> (Option<gst::Allocator>, gst::AllocationParams) {
         unsafe {
             let mut allocator = ptr::null_mut();
-            let mut params = mem::zeroed();
+            let mut params = mem::MaybeUninit::uninit();
             ffi::gst_audio_decoder_get_allocator(
                 self.as_ref().to_glib_none().0,
                 &mut allocator,
-                &mut params,
+                params.as_mut_ptr(),
             );
-            (from_glib_full(allocator), params.into())
+            (from_glib_full(allocator), params.assume_init().into())
         }
     }
 
+    #[allow(clippy::too_many_arguments)]
     fn error<T: gst::MessageErrorDomain>(
         &self,
         weight: i32,
@@ -191,46 +126,32 @@ impl<O: IsA<AudioDecoder>> AudioDecoderExtManual for O {
             &*(&elt.srcpad as *const *mut gst::ffi::GstPad as *const gst::Pad)
         }
     }
+
+    fn input_segment(&self) -> gst::Segment {
+        unsafe {
+            let ptr: &ffi::GstAudioDecoder = &*(self.as_ptr() as *const _);
+            glib::ffi::g_rec_mutex_lock(mut_override(&ptr.stream_lock));
+            let segment = ptr.input_segment;
+            glib::ffi::g_rec_mutex_unlock(mut_override(&ptr.stream_lock));
+            from_glib_none(&segment as *const gst::ffi::GstSegment)
+        }
+    }
+
+    fn output_segment(&self) -> gst::Segment {
+        unsafe {
+            let ptr: &ffi::GstAudioDecoder = &*(self.as_ptr() as *const _);
+            glib::ffi::g_rec_mutex_lock(mut_override(&ptr.stream_lock));
+            let segment = ptr.output_segment;
+            glib::ffi::g_rec_mutex_unlock(mut_override(&ptr.stream_lock));
+            from_glib_none(&segment as *const gst::ffi::GstSegment)
+        }
+    }
 }
 
+impl<O: IsA<AudioDecoder>> AudioDecoderExtManual for O {}
+
 #[macro_export]
 macro_rules! audio_decoder_error(
-    ($obj:expr, $weight:expr, $err:expr, ($msg:expr), [$debug:expr]) => { {
-        use $crate::prelude::AudioDecoderExtManual;
-        $obj.error(
-            $weight,
-            $err,
-            Some($msg),
-            Some($debug),
-            file!(),
-            module_path!(),
-            line!(),
-        )
-    }};
-    ($obj:expr, $weight:expr, $err:expr, ($msg:expr)) => { {
-        use $crate::prelude::AudioDecoderExtManual;
-        $obj.error(
-            $weight,
-            $err,
-            Some($msg),
-            None,
-            file!(),
-            module_path!(),
-            line!(),
-        )
-    }};
-    ($obj:expr, $weight:expr, $err:expr, [$debug:expr]) => { {
-        use $crate::prelude::AudioDecoderExtManual;
-        $obj.error(
-            $weight,
-            $err,
-            None,
-            Some($debug),
-            file!(),
-            module_path!(),
-            line!(),
-        )
-    }};
     ($obj:expr, $weight:expr, $err:expr, ($($msg:tt)*), [$($debug:tt)*]) => { {
         use $crate::prelude::AudioDecoderExtManual;
         $obj.error(
@@ -239,7 +160,7 @@ macro_rules! audio_decoder_error(
             Some(&format!($($msg)*)),
             Some(&format!($($debug)*)),
             file!(),
-            module_path!(),
+            $crate::glib::function_name!(),
             line!(),
         )
     }};
@@ -251,7 +172,7 @@ macro_rules! audio_decoder_error(
             Some(&format!($($msg)*)),
             None,
             file!(),
-            module_path!(),
+            $crate::glib::function_name!(),
             line!(),
         )
     }};
@@ -263,7 +184,7 @@ macro_rules! audio_decoder_error(
             None,
             Some(&format!($($debug)*)),
             file!(),
-            module_path!(),
+            $crate::glib::function_name!(),
             line!(),
         )
     }};
